@@ -1,5 +1,5 @@
 /* 
- * Copyright (C) 2006, 2007 OpenedHand Ltd.
+ * Copyright (C) 2006, 2007, 2008 OpenedHand Ltd.
  *
  * Author: Jorn Baayen <jorn@openedhand.com>
  *
@@ -58,7 +58,8 @@ struct _GSSDPClientPrivate {
 
         char              *server_id;
 
-        GSSDPSocketSource *socket_source;
+        GSSDPSocketSource *request_socket;
+        GSSDPSocketSource *multicast_socket;
 };
 
 enum {
@@ -82,7 +83,9 @@ gssdp_client_set_main_context (GSSDPClient  *client,
 static char *
 make_server_id                (void);
 static gboolean
-socket_source_cb              (gpointer      user_data);
+request_socket_source_cb      (gpointer      user_data);
+static gboolean
+multicast_socket_source_cb    (gpointer      user_data);
 
 static void
 gssdp_client_init (GSSDPClient *client)
@@ -95,13 +98,25 @@ gssdp_client_init (GSSDPClient *client)
         /* Generate default server ID */
         client->priv->server_id = make_server_id ();
 
-        /* Set up socket (Will set errno if it failed) */
-        client->priv->socket_source = gssdp_socket_source_new ();
-        if (client->priv->socket_source != NULL) {
-                g_source_set_callback ((GSource *) client->priv->socket_source,
-                                       socket_source_cb,
-                                       client,
-                                       NULL);
+        /* Set up sockets (Will set errno if it failed) */
+        client->priv->request_socket =
+                gssdp_socket_source_new (GSSPP_SOCKET_SOURCE_TYPE_REQUEST);
+        if (client->priv->request_socket != NULL) {
+                g_source_set_callback
+                        ((GSource *) client->priv->request_socket,
+                         request_socket_source_cb,
+                         client,
+                         NULL);
+        }
+
+        client->priv->multicast_socket =
+                gssdp_socket_source_new (GSSDP_SOCKET_SOURCE_TYPE_MULTICAST);
+        if (client->priv->multicast_socket != NULL) {
+                g_source_set_callback
+                        ((GSource *) client->priv->multicast_socket,
+                         multicast_socket_source_cb,
+                         client,
+                         NULL);
         }
 }
 
@@ -153,15 +168,16 @@ gssdp_client_set_property (GObject      *object,
                                                g_value_get_pointer (value));
                 break;
         case PROP_ERROR:
-                if (!client->priv->socket_source) {
+                if (!client->priv->request_socket ||
+                    !client->priv->multicast_socket) {
                         GError **error;
 
                         error = g_value_get_pointer (value);
 
-                        g_set_error (error,
-                                     GSSDP_ERROR,
-                                     GSSDP_ERROR_FAILED,
-                                     strerror (errno));
+                        g_set_error_literal (error,
+                                             GSSDP_ERROR,
+                                             GSSDP_ERROR_FAILED,
+                                             strerror (errno));
                 }
 
                 break;
@@ -178,10 +194,15 @@ gssdp_client_dispose (GObject *object)
 
         client = GSSDP_CLIENT (object);
 
-        /* Destroy the SocketSource */
-        if (client->priv->socket_source) {
-                g_source_destroy ((GSource *) client->priv->socket_source);
-                client->priv->socket_source = NULL;
+        /* Destroy the SocketSources */
+        if (client->priv->request_socket) {
+                g_source_destroy ((GSource *) client->priv->request_socket);
+                client->priv->request_socket = NULL;
+        }
+
+        if (client->priv->multicast_socket) {
+                g_source_destroy ((GSource *) client->priv->multicast_socket);
+                client->priv->multicast_socket = NULL;
         }
 
         /* Unref the context */
@@ -280,17 +301,18 @@ gssdp_client_class_init (GSSDPClientClass *klass)
                               G_SIGNAL_RUN_LAST,
                               0,
                               NULL, NULL,
-                              gssdp_marshal_VOID__STRING_INT_POINTER,
+                              gssdp_marshal_VOID__STRING_UINT_INT_POINTER,
                               G_TYPE_NONE,
-                              3,
+                              4,
                               G_TYPE_STRING,
+                              G_TYPE_UINT,
                               G_TYPE_INT,
                               G_TYPE_POINTER);
 }
 
 /**
  * gssdp_client_new
- * @main_context: The #GMainContext to associate with
+ * @main_context: The #GMainContext to associate with, or NULL
  * @error: Location to store error, or NULL
  *
  * Return value: A new #GSSDPClient object.
@@ -319,10 +341,16 @@ gssdp_client_set_main_context (GSSDPClient  *client,
 
         /* A NULL main_context is fine */
         
-        if (client->priv->socket_source) {
-                g_source_attach ((GSource *) client->priv->socket_source,
+        if (client->priv->request_socket) {
+                g_source_attach ((GSource *) client->priv->request_socket,
                                  client->priv->main_context);
-                g_source_unref ((GSource *) client->priv->socket_source);
+                g_source_unref ((GSource *) client->priv->request_socket);
+        }
+
+        if (client->priv->multicast_socket) {
+                g_source_attach ((GSource *) client->priv->multicast_socket,
+                                 client->priv->main_context);
+                g_source_unref ((GSource *) client->priv->multicast_socket);
         }
 
         g_object_notify (G_OBJECT (client), "main-context");
@@ -332,7 +360,7 @@ gssdp_client_set_main_context (GSSDPClient  *client,
  * gssdp_client_get_main_context
  * @client: A #GSSDPClient
  *
- * Return value: The #GMainContext @client is associated with.
+ * Return value: The #GMainContext @client is associated with, or NULL.
  **/
 GMainContext *
 gssdp_client_get_main_context (GSSDPClient *client)
@@ -384,6 +412,7 @@ gssdp_client_get_server_id (GSSDPClient *client)
  * _gssdp_client_send_message
  * @client: A #GSSDPClient
  * @dest_ip: The destination IP address, or NULL to broadcast
+ * @dest_port: The destination port, or NULL for default
  * @message: The message to send
  *
  * Sends @message to @dest_ip.
@@ -391,6 +420,7 @@ gssdp_client_get_server_id (GSSDPClient *client)
 void
 _gssdp_client_send_message (GSSDPClient *client,
                             const char  *dest_ip,
+                            gushort      dest_port,
                             const char  *message)
 {
         struct sockaddr_in addr;
@@ -403,12 +433,16 @@ _gssdp_client_send_message (GSSDPClient *client,
         if (dest_ip == NULL)
                 dest_ip = SSDP_ADDR;
 
-        socket_fd = gssdp_socket_source_get_fd (client->priv->socket_source);
+        /* Use default port if no port was explicitly specified */
+        if (dest_port == 0)
+                dest_port = SSDP_PORT;
+
+        socket_fd = gssdp_socket_source_get_fd (client->priv->request_socket);
 
         memset (&addr, 0, sizeof (addr));
 
         addr.sin_family      = AF_INET;
-        addr.sin_port        = htons (SSDP_PORT);
+        addr.sin_port        = htons (dest_port);
         addr.sin_addr.s_addr = inet_addr (dest_ip);
 
         res = sendto (socket_fd,
@@ -440,67 +474,89 @@ make_server_id (void)
                                 VERSION);
 }
 
-/**
- * Free a GSList of strings
- **/
-static void
-string_list_free (gpointer ptr)
+static gboolean
+parse_http_request (char                *buf,
+                    int                  len,
+                    SoupMessageHeaders **headers,
+                    int                 *type)
 {
-        GSList *list;
+        char *req_method;
 
-        list = ptr;
+        *headers = soup_message_headers_new (SOUP_MESSAGE_HEADERS_REQUEST);
 
-        while (list) {
-                g_free (list->data);
-                list = g_slist_delete_link (list, list);
+        if (soup_headers_parse_request (buf,
+                                        len,
+                                        *headers,
+                                        &req_method,
+                                        NULL,
+                                        NULL) == SOUP_STATUS_OK) {
+                if (g_ascii_strncasecmp (req_method,
+                                         SSDP_SEARCH_METHOD,
+                                         strlen (SSDP_SEARCH_METHOD)) == 0)
+                        *type = _GSSDP_DISCOVERY_REQUEST;
+                else if (g_ascii_strncasecmp (req_method,
+                                              GENA_NOTIFY_METHOD,
+                                              strlen (GENA_NOTIFY_METHOD)) == 0)
+                        *type = _GSSDP_ANNOUNCEMENT;
+                else
+                        g_warning ("Unhandled method '%s'", req_method);
+
+                g_free (req_method);
+
+                return TRUE;
+        } else {
+                soup_message_headers_free (*headers);
+                *headers = NULL;
+
+                return FALSE;
         }
 }
 
-/**
- * HTTP/1.1 headers needs to be case-insensitive and so should be our
- * hash-table of HTTP headers.
- **/
-
-/**
- * Calculates the hash for our case-insensitive Hashtable.
- **/
-static guint 
-header_hash (const char *header)
-{
-        return g_ascii_toupper (header[0]);
-}
-
-/**
- * Compares keys for our case-insensitive Hashtable.
- **/
 static gboolean
-header_equal (const gchar *header1,
-              const gchar *header2)
+parse_http_response (char                *buf,
+                    int                  len,
+                    SoupMessageHeaders **headers,
+                    int                 *type)
 {
-        return (g_ascii_strcasecmp (header1, header2) == 0);
+        guint status_code;
+
+        *headers = soup_message_headers_new (SOUP_MESSAGE_HEADERS_RESPONSE);
+
+        if (soup_headers_parse_response (buf,
+                                         len,
+                                         *headers,
+                                         NULL,
+                                         &status_code,
+                                         NULL)) {
+                if (status_code == 200)
+                        *type = _GSSDP_DISCOVERY_RESPONSE;
+                else
+                        g_warning ("Unhandled status code '%d'", status_code);
+
+                return TRUE;
+        } else {
+                soup_message_headers_free (*headers);
+                *headers = NULL;
+
+                return FALSE;
+        }
 }
 
 /**
  * Called when data can be read from the socket
  **/
-
 static gboolean
-socket_source_cb (gpointer user_data)
+socket_source_cb (GSSDPSocketSource *socket, GSSDPClient *client)
 {
-        GSSDPClient *client;
         int fd, type, len;
         size_t bytes;
         char buf[BUF_SIZE], *end;
         struct sockaddr_in addr;
         socklen_t addr_size;
-        GHashTable *hash;
-        char *req_method;
-        guint status_code;
-
-        client = GSSDP_CLIENT (user_data);
+        SoupMessageHeaders *headers;
 
         /* Get FD */
-        fd = gssdp_socket_source_get_fd (client->priv->socket_source);
+        fd = gssdp_socket_source_get_fd (socket);
 
         /* Read data */
         addr_size = sizeof (addr);
@@ -513,9 +569,9 @@ socket_source_cb (gpointer user_data)
                           &addr_size);
 
         if (bytes >= BUF_SIZE) {
-                g_warning ("Received packet of %d bytes, but the maximum "
+                g_warning ("Received packet of %u bytes, but the maximum "
                            "buffer size is %d. Packed dropped.",
-                           bytes, BUF_SIZE);
+                           (unsigned int) bytes, BUF_SIZE);
 
                 return TRUE;
         }
@@ -536,42 +592,19 @@ socket_source_cb (gpointer user_data)
         
         /* Parse message */
         type = -1;
+        headers = NULL;
 
-        hash = g_hash_table_new_full ((GHashFunc) header_hash,
-                                      (GEqualFunc) header_equal,
-                                      g_free,
-                                      string_list_free);
-
-        if (soup_headers_parse_request (buf,
-                                        len,
-                                        hash,
-                                        &req_method,
-                                        NULL,
-                                        NULL)) {
-                if (g_ascii_strncasecmp (req_method,
-                                         SSDP_SEARCH_METHOD,
-                                         strlen (SSDP_SEARCH_METHOD)) == 0)
-                        type = _GSSDP_DISCOVERY_REQUEST;
-                else if (g_ascii_strncasecmp (req_method,
-                                              GENA_NOTIFY_METHOD,
-                                              strlen (GENA_NOTIFY_METHOD)) == 0)
-                        type = _GSSDP_ANNOUNCEMENT;
-                else
-                        g_warning ("Unhandled method '%s'", req_method);
-
-                g_free (req_method);
-        } else if (soup_headers_parse_response (buf,
-                                                len,
-                                                hash,
-                                                NULL,
-                                                &status_code,
-                                                NULL)) {
-                if (status_code == 200)
-                        type = _GSSDP_DISCOVERY_RESPONSE;
-                else
-                        g_warning ("Unhandled status code '%d'", status_code);
-        } else
-                g_warning ("Unhandled message '%s'", buf);
+        if (!parse_http_request (buf,
+                                 len,
+                                 &headers,
+                                 &type)) {
+                if (!parse_http_response (buf,
+                                          len,
+                                          &headers,
+                                          &type)) {
+                        g_warning ("Unhandled message '%s'", buf);
+                }
+        }
         
         /* Emit signal if parsing succeeded */
         if (type >= 0) {
@@ -579,11 +612,33 @@ socket_source_cb (gpointer user_data)
                                signals[MESSAGE_RECEIVED],
                                0,
                                inet_ntoa (addr.sin_addr),
+                               ntohs (addr.sin_port),
                                type,
-                               hash);
+                               headers);
         }
 
-        g_hash_table_unref (hash);
+        if (headers)
+                soup_message_headers_free (headers);
 
         return TRUE;
 }
+
+static gboolean
+request_socket_source_cb (gpointer user_data)
+{
+        GSSDPClient *client;
+
+        client = GSSDP_CLIENT (user_data);
+
+        return socket_source_cb (client->priv->request_socket, client);
+}
+
+static gboolean
+multicast_socket_source_cb (gpointer user_data)
+{
+        GSSDPClient *client;
+
+        client = GSSDP_CLIENT (user_data);
+
+        return socket_source_cb (client->priv->multicast_socket, client);
+}
